// We have to include InitGuid.h first so that our property IDs will be defined.
#include <InitGuid.h> 

#include "CaptureStream.h"
#include "DataBuffer.h"
#include "ReferenceTimeUtils.h"

#include <stdexcept>

CaptureStream::CaptureStream(IMMDevicePtr &spDevice) : m_hEvent(NULL)
{
    if (!spDevice)
        throw std::runtime_error("Null device. Can't create audio stream.");

    // Activating the device, which gives us the audio client.
    IAudioClientPtr spAudioClient;
    if (FAILED(spDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL, 
        reinterpret_cast<void **>(&spAudioClient))))
    {
        throw std::runtime_error("Failed to acquire default audio client.");
    }

    // Lets find out what format the device is currently using. This should be 
    // a valid format to use.
    IPropertyStorePtr spPropStore;
    if (FAILED(spDevice->OpenPropertyStore(STGM_READ, &spPropStore)))
    {
        throw std::runtime_error("Couldn't open property store for device!");
    }
    PROPVARIANT varAudioFormat;
    PropVariantInit(&varAudioFormat);
    if (FAILED(spPropStore->GetValue(PKEY_AudioEngine_DeviceFormat, &varAudioFormat)))
    {
        throw std::runtime_error("Couldn't determine current format of the device.");
    }

    std::shared_ptr<AudioFormat> spAudioFormat = std::make_shared<AudioFormat>(
        reinterpret_cast<WAVEFORMATEX *>(varAudioFormat.blob.pBlobData));

    // Determining the audio device's minimum processing period, which helps us
    // choose a reasonable buffer size.
    REFERENCE_TIME audioDeviceMinPeriodInRefTimes = 0;
    if (FAILED(spAudioClient->GetDevicePeriod(NULL, &audioDeviceMinPeriodInRefTimes)))
    {
        throw std::runtime_error("Couldn't determine device period.");
    }

    // For minimum latency, we could use the device period to allocate our 
    // buffer size. However, this sometimes leads to stutter, pops, and other 
    // artifacts in the rendered audio. We don't care about minor latency here.
    // So we'll double this buffer length in an effort to produce cleaner audio.
    // If we still hear artifacts in the audio, this can be tweaked. 
    REFERENCE_TIME bufferLengthInRefTimes = audioDeviceMinPeriodInRefTimes * 2;

    // Allocates two buffers
    HRESULT hr = 0;
    if (FAILED(hr = spAudioClient->Initialize(AUDCLNT_SHAREMODE_EXCLUSIVE, 
        AUDCLNT_STREAMFLAGS_EVENTCALLBACK, bufferLengthInRefTimes, 
        bufferLengthInRefTimes, spAudioFormat->getWaveFormat(), NULL)))
    {
        // Can occur in Windows 7 and later.
        if (hr == AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED)
            bufferLengthInRefTimes = reinitializeStreamWithAlignedBuffer(spDevice);
        else
            throw std::runtime_error("Could not initialize stream.");
    }

    // Setting up the event handle.
    m_hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (m_hEvent == NULL)
    {
        throw std::runtime_error("Could not create event.");
    }

    if (FAILED(spAudioClient->SetEventHandle(m_hEvent)))
    {
        throw std::runtime_error("Could not set event on audio client.");
    }

    // Obtaining capture client.
    IAudioCaptureClientPtr spCaptureClient;
    if (FAILED(spAudioClient->GetService(__uuidof(IAudioCaptureClient),
        reinterpret_cast<void **>(&spCaptureClient))))
    {
        throw std::runtime_error("Failed to acquire capture client.");
    }

    m_spAudioClient = spAudioClient;
    m_spCaptureClient = spCaptureClient;
    m_spAudioFormat = spAudioFormat;
}

CaptureStream::~CaptureStream()
{
    if (m_hEvent)
        CloseHandle(m_hEvent);
}

// Called from the capture thread. Performs a wait appropriate for our 
// buffering strategy. In this case, we wait for an event to be set.
bool
CaptureStream::waitOnBuffer()
{
    if (WaitForSingleObject(m_hEvent, 500) != WAIT_OBJECT_0)
        return false;

    return true;
}

void 
CaptureStream::captureNextFrames(DataBufferWriter &dataBuffer)
{        
    // Note that GetNextPacketSize() cannot be used with exclusive-mode streams.
    
    DWORD flags = 0;
    BYTE *pData = nullptr;
    UINT32 numFramesRead = 0;

    if (FAILED(m_spCaptureClient->GetBuffer(&pData, &numFramesRead, &flags, NULL, NULL)))
        throw std::runtime_error("Couldn't obtain buffer.");

    UINT32 bytesRead = numFramesRead * m_spAudioFormat->getNumChannels() *
        m_spAudioFormat->getBytesPerSample();

    if ((flags & AUDCLNT_BUFFERFLAGS_SILENT) != 0)
        dataBuffer.writeValue(0, bytesRead);
    else
        dataBuffer.writeBytes(pData, bytesRead);

    m_spCaptureClient->ReleaseBuffer(numFramesRead);
}

REFERENCE_TIME
CaptureStream::reinitializeStreamWithAlignedBuffer(IMMDevicePtr &spDevice)
{
    if (!m_spAudioClient)
    {
        throw std::runtime_error("Can't reinitialize a NULL audio client.");
    }

    // Retrieves the next highest buffer size, in frames.
    UINT32 bufferLengthInFrames = 0;
    if (FAILED(m_spAudioClient->GetBufferSize(&bufferLengthInFrames)))
    {
        throw std::runtime_error("Could not determine appropriate buffer size.");
    }
    m_spAudioClient.Release();

// Calculating an aligned buffer size in REFERENCE_TIME units, per MSDN.
// https://msdn.microsoft.com/en-us/library/windows/desktop/dd370875%28v=vs.85%29.aspx

    REFERENCE_TIME bufferLengthInRefTimes = 
        (REFERENCE_TIME)((10000.0 * 1000 / 
        m_spAudioFormat->getSampleRateInHz() * bufferLengthInFrames) + 0.5);

    // Reactivating the device, which gives us the audio client.
    if (FAILED(spDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL,
        reinterpret_cast<void **>(&m_spAudioClient))))
    {
        throw std::runtime_error("Failed to acquire default audio client.");
    }

    if (FAILED(m_spAudioClient->Initialize(AUDCLNT_SHAREMODE_EXCLUSIVE, 
        AUDCLNT_STREAMFLAGS_EVENTCALLBACK, bufferLengthInRefTimes, 
        bufferLengthInRefTimes, m_spAudioFormat->getWaveFormat(), NULL)))
    {
        throw std::runtime_error(
            "Failed to initialize audio client with aligned buffer size.");
    }

    return bufferLengthInRefTimes;
}
