#include "AppStrings.h"
#include "AudioFormat.h"
#include "KeyMonitor.h"
#include "CaptureStream.h"
#include "CaptureThread.h"
#include "AudioDevices.h"
#include "DataBuffer.h"
#include "RenderStream.h"
#include "RenderThread.h"

#include <iostream>
#include <memory>

/**
 * Prints out the audio format being used.
 */
void 
printAudioFormat(const AudioFormat &audioFormat)
{
    std::cout << "Device: " << audioFormat.getSampleRateInHz() << " Hz, ";
    switch (audioFormat.getEncoding())
    {
        case AudioEncoding::FLOATING_POINT:
            std::cout << "32-Bit Floating Point"; break;
        case AudioEncoding::PCM_8:
            std::cout << "8-Bit PCM"; break;
        case AudioEncoding::PCM_16:
            std::cout << "16-Bit PCM"; break;
        case AudioEncoding::PCM_24:
            std::cout << "24-Bit PCM"; break;
        case AudioEncoding::PCM_24_IN_32:
            std::cout << "24-Bit PCM (32-bit Container)"; break;
        case AudioEncoding::PCM_32:
            std::cout << "PCM_32 PCM"; break;
    }
    std::cout << std::endl << std::endl;
}

/**
 * @param audioFormat The audio format.
 * @param durationInSeconds How many seconds the buffer should be sized for.
 */
uint32_t
calculateOptimalBufferSizeInBytes(const AudioFormat &audioFormat, 
    uint32_t durationInSeconds)
{
    uint32_t channels = audioFormat.getNumChannels();
    uint32_t samplesPerSec = audioFormat.getSampleRateInHz();
    uint32_t bytesPerSample = audioFormat.getBytesPerSample();

    uint32_t frameSizeInBytes = channels * bytesPerSample;
    uint32_t oneSecondOfAudioSizeInBytes = frameSizeInBytes * samplesPerSec;
    return oneSecondOfAudioSizeInBytes * durationInSeconds;
}

int 
main(int argc, const char * argv[])
{
    SetConsoleTitleA(CONSOLE_TITLE);

    std::ios::sync_with_stdio();

    // Initializing COM. This is required for all applications making use
    // of COM components.
    HRESULT hr = CoInitializeEx(NULL, COINIT_APARTMENTTHREADED);

    std::cout << "Capturing Stream Demo using " MODE_TITLE << std::endl
              << "Recording for 30 seconds or until 'q' is pressed."
              << std::endl << std::endl;

    KeyMonitor keyMonitor;

    try
    {
        // Creating capture stream.
        IMMDevicePtr spInDevice = AudioDevices::getDefaultCaptureDevice();
        CaptureStream inStream(spInDevice);
        printAudioFormat(*inStream.getAudioFormat());

        // Creating our buffer.
        uint32_t bufferSizeInBytes = calculateOptimalBufferSizeInBytes(
            *(inStream.getAudioFormat()), 30);
        std::unique_ptr<unsigned char[]> spBuffer(new unsigned char[bufferSizeInBytes]);
        
        // Capturing audio.
        DataBufferWriter bufferWriter(spBuffer.get(), bufferSizeInBytes);
        CaptureThread captureThread(inStream, bufferWriter);
        captureThread.start();

        while (!captureThread.isFinished())
        {            
            KeyMonitor::KeyPress keyPress = keyMonitor.getKeyPress(false);
            if (keyPress == KeyMonitor::KeyPress::Q)
                break;
            Sleep(100);
        }
        captureThread.stop();

        // Playing back audio.
        std::cout << "Playing back audio until finished or "
            "until 'q' is pressed. " << std::endl;
        IMMDevicePtr spOutDevice = AudioDevices::getDefaultRenderDevice();
        RenderStream outStream(spOutDevice, inStream.getAudioFormat());

        // Rendering audio.
        DataBufferReader bufferReader(spBuffer.get(), bufferWriter.getBufferUsed());
        RenderThread renderThread(outStream, bufferReader);
        renderThread.start();

        while (!bufferReader.isEmpty())
        {
            KeyMonitor::KeyPress keyPress = keyMonitor.getKeyPress(false);
            if (keyPress == KeyMonitor::KeyPress::Q)
                break;
            Sleep(100);
        }

        renderThread.stop();
        std::cout << "Done." << std::endl;
    }
    catch (const std::runtime_error &e)
    {
        std::cerr << e.what() << std::endl;
    }    

    CoUninitialize();

    return 0;
}

