#include "CaptureStream.h"
#include "DataBuffer.h"
#include "ReferenceTimeUtils.h"

#include <stdexcept>

CaptureStream::CaptureStream(IMMDevicePtr &spDevice) : m_hEvent(NULL)
{
    if (!spDevice)
        throw std::runtime_error("Null device. Can't create audio stream.");

    // Activating the device, which gives us the audio client.
    IAudioClientPtr spAudioClient;
    if (FAILED(spDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL,
        reinterpret_cast<void **>(&spAudioClient))))
    {
        throw std::runtime_error("Failed to acquire default audio client.");
    }

    // Determining current shared-mode mix format.
    WAVEFORMATEX *pWaveFormat = nullptr;
    if (FAILED(spAudioClient->GetMixFormat(&pWaveFormat)))
    {
        throw std::runtime_error("Failed to determine mix format.");
    }

    std::shared_ptr<AudioFormat> spAudioFormat = 
        std::make_shared<AudioFormat>(pWaveFormat);
    
    // Initializing audio client.
    if (FAILED(spAudioClient->Initialize(AUDCLNT_SHAREMODE_SHARED, 
        AUDCLNT_STREAMFLAGS_EVENTCALLBACK, 0, 0, pWaveFormat, NULL)))
    {
        throw std::runtime_error("Could not initialize audio client with shared mode.");
    }

    // Setting up the event handle.
    m_hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (m_hEvent == NULL)
    {
        throw std::runtime_error("Could not create event.");
    }

    if (FAILED(spAudioClient->SetEventHandle(m_hEvent)))
    {
        throw std::runtime_error("Could not set event on audio client.");
    }

    // Obtaining capture client.
    IAudioCaptureClientPtr spCaptureClient;
    if (FAILED(spAudioClient->GetService(__uuidof(IAudioCaptureClient),
        reinterpret_cast<void **>(&spCaptureClient))))
    {
        throw std::runtime_error("Failed to acquire capture client.");
    }

    m_spAudioClient = spAudioClient;
    m_spCaptureClient = spCaptureClient;
    m_spAudioFormat = spAudioFormat;
}

CaptureStream::~CaptureStream()
{
    if (m_hEvent)
        CloseHandle(m_hEvent);
}

// Called from the render thread. Performs a wait appropriate for our 
// buffering strategy. In this case, we wait for an event to be set.
bool
CaptureStream::waitOnBuffer()
{
    if (WaitForSingleObject(m_hEvent, 500) != WAIT_OBJECT_0)
        return false;

    return true;
}

void 
CaptureStream::captureNextFrames(DataBufferWriter &dataBuffer)
{        
    UINT32 nextPacketSize = 0;

    if (FAILED(m_spCaptureClient->GetNextPacketSize(&nextPacketSize)))
        throw std::runtime_error("Couldn't read next packet size.");

    while (nextPacketSize > 0)
    {
        DWORD flags = 0;
        BYTE *pData = nullptr;
        UINT32 numFramesRead = 0;

        if (FAILED(m_spCaptureClient->GetBuffer(&pData, &numFramesRead, 
            &flags, NULL, NULL)))
            throw std::runtime_error("Couldn't obtain buffer.");

        UINT32 bytesRead = numFramesRead * m_spAudioFormat->getNumChannels() *
            m_spAudioFormat->getBytesPerSample();

        if ((flags & AUDCLNT_BUFFERFLAGS_SILENT) != 0)
            dataBuffer.writeValue(0, bytesRead);
        else
            dataBuffer.writeBytes(pData, bytesRead);

        m_spCaptureClient->ReleaseBuffer(numFramesRead);

        if (FAILED(m_spCaptureClient->GetNextPacketSize(&nextPacketSize)))
            throw std::runtime_error("Couldn't read next packet size.");
    }
}

