#include "AudioFormat.h"

#include <Mmreg.h>
#include <stdexcept>

AudioFormat::AudioFormat(WAVEFORMATEX *pWaveFormat) : 
    m_pWaveFormat(nullptr), m_encoding(AudioEncoding::UNSUPPORTED)
{
    if (pWaveFormat == nullptr)
        throw std::runtime_error("WAVEFORMATEX cannot be null.");
    
    m_encoding = determineAudioEncoding(*pWaveFormat);
    if (m_encoding == AudioEncoding::UNSUPPORTED)
    {
        CoTaskMemFree(pWaveFormat);
        throw std::runtime_error("Unsupported audio format.");
    }
    m_pWaveFormat = pWaveFormat;
}

// Destructor.
AudioFormat::~AudioFormat()
{
    if (m_pWaveFormat)
        CoTaskMemFree(m_pWaveFormat);
}

/**
 * We only support floating point and 8/16/24/32 bit PCM formats.
 */
AudioEncoding
AudioFormat::determineAudioEncoding(const WAVEFORMATEX &format)
{
    if (format.wFormatTag == WAVE_FORMAT_IEEE_FLOAT)
    {
        if (format.wBitsPerSample == (sizeof(float) * 8))
            return AudioEncoding::FLOATING_POINT;
        else
            return AudioEncoding::UNSUPPORTED;
    }
    else
    if (format.wFormatTag == WAVE_FORMAT_PCM)
    {
        switch (format.wBitsPerSample)
        {
            // All bits must be considered valid.
            case 8  : return AudioEncoding::PCM_8;
            case 16 : return AudioEncoding::PCM_16;
            case 24 : return AudioEncoding::PCM_24;
            case 32 : return AudioEncoding::PCM_32;
            default : return AudioEncoding::UNSUPPORTED;
        }
    }
    else
    if ((format.wFormatTag == WAVE_FORMAT_EXTENSIBLE) && (format.cbSize >= 22))
    {
        const WAVEFORMATEXTENSIBLE *pFormatExtensible =
            reinterpret_cast<const WAVEFORMATEXTENSIBLE *>(&format);
        if (pFormatExtensible->SubFormat == KSDATAFORMAT_SUBTYPE_IEEE_FLOAT)
        {
            if (format.wBitsPerSample == (sizeof(float) * 8))
                return AudioEncoding::FLOATING_POINT;
            else
                return AudioEncoding::UNSUPPORTED;
        }
        else
        if (pFormatExtensible->SubFormat == KSDATAFORMAT_SUBTYPE_PCM)
        {
            switch (pFormatExtensible->Samples.wValidBitsPerSample)
            {
                case 8:  
                { 
                    return (format.wBitsPerSample == 8) ? 
                        AudioEncoding::PCM_8 : AudioEncoding::UNSUPPORTED; 
                }
                case 16: 
                { 
                    return (format.wBitsPerSample == 16) ? 
                        AudioEncoding::PCM_16 : AudioEncoding::UNSUPPORTED; 
                }
                case 32: 
                { 
                    return (format.wBitsPerSample == 32) ? 
                        AudioEncoding::PCM_32 : AudioEncoding::UNSUPPORTED; 
                }
                case 24:
                {
                    if (format.wBitsPerSample == 32)
                        return AudioEncoding::PCM_24_IN_32;
                    else
                    if (format.wBitsPerSample == 24)
                        return AudioEncoding::PCM_24;
                    else
                        return AudioEncoding::UNSUPPORTED;
                }
            }
        }
    }

    return AudioEncoding::UNSUPPORTED;
}
