#include "CaptureStream.h"
#include "DataBuffer.h"
#include "ReferenceTimeUtils.h"

#include <stdexcept>

CaptureStream::CaptureStream(IMMDevicePtr &spDevice) : m_sleepTimeInMs(0)
{
    if (!spDevice)
        throw std::runtime_error("Null device. Can't create audio stream.");

    // Activating the device, which gives us the audio client.
    IAudioClientPtr spAudioClient;
    if (FAILED(spDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL,
        reinterpret_cast<void **>(&spAudioClient))))
    {
        throw std::runtime_error("Failed to acquire default audio client.");
    }

    // Determining current shared-mode mix format.
    WAVEFORMATEX *pWaveFormat = nullptr;
    if (FAILED(spAudioClient->GetMixFormat(&pWaveFormat)))
    {
        throw std::runtime_error("Failed to determine mix format.");
    }

    std::shared_ptr<AudioFormat> spAudioFormat = 
        std::make_shared<AudioFormat>(pWaveFormat);
    
    // Determining the audio engine's processing period, which helps us
    // choose a reasonable buffer size.
    REFERENCE_TIME audioEnginePeriodInRefTimes = 0;
    if (FAILED(spAudioClient->GetDevicePeriod(&audioEnginePeriodInRefTimes, NULL)))
    {
        throw std::runtime_error("Couldn't determine device period.");
    }
    REFERENCE_TIME bufferLengthInRefTimes = audioEnginePeriodInRefTimes * 2;

    // Initializing audio client.
    if (FAILED(spAudioClient->Initialize(AUDCLNT_SHAREMODE_SHARED, 0,
        bufferLengthInRefTimes, 0, pWaveFormat, NULL)))
    { 
        throw std::runtime_error("Could not initialize audio client with shared mode.");
    }

    // Obtaining capture client.
    IAudioCaptureClientPtr spCaptureClient;
    if (FAILED(spAudioClient->GetService(__uuidof(IAudioCaptureClient),
        reinterpret_cast<void **>(&spCaptureClient))))
    {
        throw std::runtime_error("Failed to acquire capture client.");
    }

    UINT32 bufferSizeInFrames = 0;
    // Getting total number of frames, which may be different than our ask.
    if (FAILED(spAudioClient->GetBufferSize(&bufferSizeInFrames)))
    {
        throw std::runtime_error("Couldn't obtain total buffer frame count.");
    }

    // Determining an appropriate time interval to sleep between
    // reading from the data buffer.
    double durationOfBufferInSeconds = static_cast<double>(bufferSizeInFrames) /
        static_cast<double>(spAudioFormat->getSampleRateInHz());
    double durationOfBufferInMillis = durationOfBufferInSeconds * 1000.0;
    m_sleepTimeInMs = (DWORD)(durationOfBufferInMillis / 2.0);

    m_spAudioClient = spAudioClient;
    m_spCaptureClient = spCaptureClient;
    m_spAudioFormat = spAudioFormat;
}

// Called from the capture thread. Performs a wait appropriate for our 
// buffering strategy. In this case, we simply sleep for a given duration
// of time. 
bool
CaptureStream::waitOnBuffer()
{
    Sleep(m_sleepTimeInMs);
    return true;
}

void 
CaptureStream::captureNextFrames(DataBufferWriter &dataBuffer)
{        
    UINT32 nextPacketSize = 0;

    if (FAILED(m_spCaptureClient->GetNextPacketSize(&nextPacketSize)))
        throw std::runtime_error("Couldn't read next packet size.");

    while (nextPacketSize > 0)
    {
        DWORD flags = 0;
        BYTE *pData = nullptr;
        UINT32 numFramesRead = 0;

        if (FAILED(m_spCaptureClient->GetBuffer(&pData, &numFramesRead, 
            &flags, NULL, NULL)))
            throw std::runtime_error("Couldn't obtain buffer.");

        UINT32 bytesRead = numFramesRead * m_spAudioFormat->getNumChannels() *
            m_spAudioFormat->getBytesPerSample();

        if ((flags & AUDCLNT_BUFFERFLAGS_SILENT) != 0)
            dataBuffer.writeValue(0, bytesRead);
        else
            dataBuffer.writeBytes(pData, bytesRead);

        m_spCaptureClient->ReleaseBuffer(numFramesRead);

        if (FAILED(m_spCaptureClient->GetNextPacketSize(&nextPacketSize)))
            throw std::runtime_error("Couldn't read next packet size.");
    }
}

