#ifndef _databuffer__h_
#define _databuffer__h_

#include <cstdint>

class AudioFormat;

/**
 * Wraps a data buffer, providing simple write operations against it. It does
 * not take ownership of the buffer. It also does not provide synchronized
 * access to the buffer.
 */
class DataBufferWriter
{

public:

    DataBufferWriter(unsigned char *pBuffer, uint32_t bufferSize) :
        m_pBuffer(pBuffer), m_bufferSize(bufferSize), m_bufferUsed(0) {}

    // Returns whether or not we've filled the buffer.
    inline bool isBufferFull() const
    {
        if (m_bufferUsed >= m_bufferSize)
            return true;

        return (m_pBuffer == nullptr) || (m_bufferUsed >= m_bufferSize);
    }

    // Returns how much of the buffer has been written into.
    inline uint32_t getBufferUsed() const 
    { 
        return m_bufferUsed; 
    }

    // Writes the specified into the buffer and returns the number of 
    // bytes written.
    uint32_t writeBytes(unsigned char *pBuffer, uint32_t bufferLen);

    // Writes the specified byte into the buffer the specified number
    // of times.
    uint32_t writeValue(unsigned char byte, uint32_t count);

private:

    unsigned char *m_pBuffer;
    uint32_t m_bufferSize;
    uint32_t m_bufferUsed;    
};

/**
 * Wraps a data buffer, providing simple read operations against it. It does
 * not take ownership of the buffer. It also does not provide synchronized 
 * access to the buffer.
 */
class DataBufferReader
{
public:

    // Constructor.
    DataBufferReader(unsigned char *pBuffer, uint32_t bufferSize) :
        m_pBuffer(pBuffer), m_bufferSize(bufferSize), m_bytesRead(0) {}
    
    // Returns whether or not we've read all of the data out of the buffer.
    inline bool isEmpty() const 
    { 
        return (m_pBuffer == nullptr) || (m_bytesRead >= m_bufferSize);
    }

    // Reads data into the specified buffer.
    uint32_t readBytes(unsigned char *pBuffer, uint32_t bufferLen);

private:

    unsigned char *m_pBuffer;
    uint32_t m_bufferSize;
    uint32_t m_bytesRead;
};

#endif

