#include "RenderStream.h"
#include "DataBuffer.h"

#include <stdexcept>

#ifndef AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM
#define AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM 0x80000000
#endif

#ifndef AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY
#define AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY 0x08000000 
#endif

RenderStream::RenderStream(IMMDevicePtr &spDevice, 
    const std::shared_ptr<AudioFormat>& spAudioFormat) :
        m_bufferSizeInFrames(0), m_hEvent(NULL)
{
    if (!spDevice)
        throw std::runtime_error("Null device. Can't create audio stream.");

    // Activating the device, which gives us the audio client.
    IAudioClientPtr spAudioClient;
    if (FAILED(spDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL, 
        reinterpret_cast<void **>(&spAudioClient))))
    {
        throw std::runtime_error("Failed to acquire default audio client.");
    }

    // Initializing audio client.
    DWORD flags = AUDCLNT_STREAMFLAGS_EVENTCALLBACK | 
                  AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM | 
                  AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;

    if (FAILED(spAudioClient->Initialize(AUDCLNT_SHAREMODE_SHARED,
        flags, 0, 0, spAudioFormat->getWaveFormat(), NULL)))
    {
        throw std::runtime_error("Could not initialize audio client with shared mode.");
    }

    // Setting up the event handle.
    m_hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (m_hEvent == NULL)
    {
        throw std::runtime_error("Could not create event.");
    }

    if (FAILED(spAudioClient->SetEventHandle(m_hEvent)))
    {
        throw std::runtime_error("Could not set event on audio client.");
    }

    // Obtaining render client.
    IAudioRenderClientPtr spRenderClient;
    if (FAILED(spAudioClient->GetService(__uuidof(IAudioRenderClient),
        reinterpret_cast<void **>(&spRenderClient))))
    {
        throw std::runtime_error("Failed to acquire render client.");
    }

    // Getting total number of frames, which may be different than our ask.
    if (FAILED(spAudioClient->GetBufferSize(&m_bufferSizeInFrames)))
    {
        throw std::runtime_error("Couldn't obtain total buffer frame count.");
    }

    m_spAudioClient = spAudioClient;
    m_spRenderClient = spRenderClient;
    m_spAudioFormat = spAudioFormat;
}

RenderStream::~RenderStream()
{
    if (m_hEvent)
        CloseHandle(m_hEvent);
}

// Called from the render thread. Performs a wait appropriate for our 
// buffering strategy. In this case, we wait for an event to be set.
bool
RenderStream::waitOnBuffer()
{
    if (WaitForSingleObject(m_hEvent, 500) != WAIT_OBJECT_0)
        return false;

    return true;
}

void
RenderStream::renderNextFrames(DataBufferReader &buffer)
{
    // See how much buffer space is available. Frame padding is the amount
    // of valid, unread data currently in the buffer. We can use this to figure
    // out how many frames are invalid and available for writing.
    UINT32 numFramesPadding = 0;
    m_spAudioClient->GetCurrentPadding(&numFramesPadding);

    UINT32 numFramesAvailable = m_bufferSizeInFrames - numFramesPadding;

    BYTE *pBuffer = nullptr;
    if (FAILED(m_spRenderClient->GetBuffer(numFramesAvailable, &pBuffer)))
        throw std::runtime_error("Couldn't obtain buffer.");

    int frameSizeInBytes = m_spAudioFormat->getFrameSizeInBytes();
    uint32_t bytesAvailable = numFramesAvailable * frameSizeInBytes;
    uint32_t bytesRead = buffer.readBytes(pBuffer, bytesAvailable);
    uint32_t framesRead = bytesRead / frameSizeInBytes;

    // Marks the frames we just wrote as valid.
    m_spRenderClient->ReleaseBuffer(framesRead, 0 /*flags*/);
}

