// We have to include InitGuid.h first so that our property IDs will be defined.
#include <InitGuid.h> 

#include "DeviceInfoStreamWriter.h"
#include "ComTypes.h"

#include <map>
#include <stdexcept>

#include <Mmreg.h>
#include <Functiondiscoverykeys_devpkey.h>

void
DeviceInfoStreamWriter::writeDeviceInfo(IMMDevice &device, std::wostream &output)
{
    output << std::endl;
    output << "===========================================================" << std::endl;
    writeDataFlow(device, output);
    output << ", ";
    writeId(device, output);
    output << std::endl;
    output << "===========================================================" << std::endl;
    writeState(device, output);
    writeProperties(device, output);
}

void
DeviceInfoStreamWriter::writeDataFlow(IMMDevice &device, std::wostream &output)
{
    IMMEndpointPtr spEndpoint;
    if (FAILED(device.QueryInterface(__uuidof(IMMEndpoint),
        reinterpret_cast<void **>(&spEndpoint))))
    {
        output << "Warning: Not an audio endpoint device." << std::endl;
        return;
    }

    EDataFlow flow = eRender;
    if (SUCCEEDED(spEndpoint->GetDataFlow(&flow)))
    {
        if (flow == eRender)
            output << "Output Device";
        else
            output << "Input Device";
    }
}

void
DeviceInfoStreamWriter::writeId(IMMDevice &device, std::wostream &output)
{
    LPWSTR pID = nullptr;
    if (SUCCEEDED(device.GetId(&pID)))
    {
        output << "ID: " << pID;
        CoTaskMemFree(pID);
    }
}

void
DeviceInfoStreamWriter::writeState(IMMDevice &device, std::wostream &output)
{
    DWORD state = 0;
    if (SUCCEEDED(device.GetState(&state)))
    {
        output << "State: ";
        switch (state)
        {
            case DEVICE_STATE_ACTIVE     : output << "Active"; break;
            case DEVICE_STATE_DISABLED   : output << "Disabled"; break;
            case DEVICE_STATE_NOTPRESENT : output << "Not Present"; break;
            case DEVICE_STATE_UNPLUGGED  : output << "Unplugged"; break;
            default                      : output << "Unknown"; break;
        }
        output << std::endl;
    }
}

void
DeviceInfoStreamWriter::writeProperties(IMMDevice &device, std::wostream &output)
{
    IPropertyStorePtr spPropStore;
    if (FAILED(device.OpenPropertyStore(STGM_READ, &spPropStore)))
    {
        throw std::runtime_error("Couldn't open property store for device!");
    }

    writeNameProperty(*spPropStore, output);
    writeFormFactoryProperty(*spPropStore, output);
    writePhysicalSpeakersProperty(*spPropStore, output);
    writeCurrentAudioFormat(*spPropStore, output);
    writeOEMAudioFormat(*spPropStore, output);
}

void
DeviceInfoStreamWriter::writeNameProperty(IPropertyStore &propertyStore, std::wostream &output)
{
    PROPVARIANT varName;
    PropVariantInit(&varName);
    if (SUCCEEDED(propertyStore.GetValue(PKEY_Device_FriendlyName, &varName)))
    {
        output << "Name: ";
        writeStringProperty(varName, output);
        output << std::endl;
    }

    PropVariantClear(&varName);
}

void
DeviceInfoStreamWriter::writeFormFactoryProperty(IPropertyStore &propertyStore, std::wostream &output)
{
    static const char * formFactors[] =
    {
        "Remote Network Device",
        "Speakers",
        "LineLevel",
        "Headphones",
        "Microphone",
        "Headset",
        "Handset",
        "Unknown Digital Passthrough",
        "SPDIF",
        "Digital Audio Display Device",
        "Unknown Form Factor",
    };

    PROPVARIANT varFormFactor;
    PropVariantInit(&varFormFactor);
    if (SUCCEEDED(propertyStore.GetValue(PKEY_AudioEndpoint_FormFactor, &varFormFactor)))
    {
        output << "Form Factor: ";
     
        if ((varFormFactor.vt == VT_EMPTY) || (varFormFactor.vt == VT_NULL))
            output << "[ NOT SET ]";
        else
        if (varFormFactor.vt != VT_UI4)
            output << "[ Error: Property not a form factor. ]";
        else
            output << formFactors[varFormFactor.uintVal];

        output << std::endl;
    }
    PropVariantClear(&varFormFactor);
}

void
DeviceInfoStreamWriter::writePhysicalSpeakersProperty(IPropertyStore &propertyStore, std::wostream &output)
{
    static const std::map<UINT, std::wstring> labelMap =
    {
        { SPEAKER_FRONT_LEFT,            L"Front Left" },
        { SPEAKER_FRONT_RIGHT,           L"Front Right" },
        { SPEAKER_FRONT_CENTER,          L"Front Center" },
        { SPEAKER_LOW_FREQUENCY,         L"Low Frequency" },
        { SPEAKER_BACK_LEFT,             L"Back Left" },
        { SPEAKER_BACK_RIGHT,            L"Back Right" },
        { SPEAKER_FRONT_LEFT_OF_CENTER,  L"Front Left of Center" },
        { SPEAKER_FRONT_RIGHT_OF_CENTER, L"Front Right of Center" },
        { SPEAKER_BACK_CENTER,           L"Back Center" },
        { SPEAKER_SIDE_LEFT,             L"Side Left" },
        { SPEAKER_SIDE_RIGHT,            L"Side Right" },
        { SPEAKER_TOP_CENTER,            L"Top Center" },
        { SPEAKER_TOP_FRONT_LEFT,        L"Front Left" },
        { SPEAKER_TOP_FRONT_CENTER,      L"Front Center" },
        { SPEAKER_TOP_FRONT_RIGHT,       L"Front Right" },
        { SPEAKER_TOP_BACK_LEFT,         L"Back Left" },
        { SPEAKER_TOP_BACK_CENTER,       L"Back Center" },
        { SPEAKER_TOP_BACK_RIGHT,        L"Back Right" }
    };
    
    PROPVARIANT varSpeakers;
    PropVariantInit(&varSpeakers);
    if (SUCCEEDED(propertyStore.GetValue(PKEY_AudioEndpoint_PhysicalSpeakers, &varSpeakers)))
    {
        output << "Speaker Configuration: ";

        if ((varSpeakers.vt == VT_EMPTY) || (varSpeakers.vt == VT_NULL))
            output << "[ NOT SET ]";
        else
        if (varSpeakers.vt != VT_UI4)
            output << "[ Error: Property not a speaker configuration. ]";
        else
        {
            std::wstring speakers;
            
            for (auto iter = labelMap.begin(); iter != labelMap.end(); ++iter)
            {
                if ((varSpeakers.uintVal & iter->first) != 0)
                {
                    if (!speakers.empty())
                        speakers.append(L", ");

                    speakers.append(iter->second);
                }
            }

            output << speakers.c_str();
        }

        output << std::endl;
    }
    PropVariantClear(&varSpeakers);
}

void
DeviceInfoStreamWriter::writeCurrentAudioFormat(IPropertyStore &propertyStore, std::wostream &output)
{
    PROPVARIANT varAudioFormat;
    PropVariantInit(&varAudioFormat);
    if (SUCCEEDED(propertyStore.GetValue(PKEY_AudioEngine_DeviceFormat, &varAudioFormat)))
    {
        output << "Current Audio Format: ";
        writeWaveFormatProperty(varAudioFormat, output);
        output << std::endl;
    }
    PropVariantClear(&varAudioFormat);
}

void
DeviceInfoStreamWriter::writeOEMAudioFormat(IPropertyStore &propertyStore, std::wostream &output)
{
    PROPVARIANT varAudioFormat;
    PropVariantInit(&varAudioFormat);
    if (SUCCEEDED(propertyStore.GetValue(PKEY_AudioEngine_OEMFormat, &varAudioFormat)))
    {
        output << "OEM Audio Format: ";
        writeWaveFormatProperty(varAudioFormat, output);
        output << std::endl;
    }
    PropVariantClear(&varAudioFormat);
}

void
DeviceInfoStreamWriter::writeStringProperty(const PROPVARIANT &property, std::wostream &output)
{
    switch (property.vt)
    {
        case VT_EMPTY  : output << "[ EMPTY ]"; break;
        case VT_NULL   : output << "[ NULL ]"; break;
        case VT_LPWSTR : output << property.pwszVal; break;
        default        : output << "[ Unsupported property type. ]"; break;
    }
}

/**
 * @param property The wave format property.
 * @return A wstring representation of the property.
 */
void
DeviceInfoStreamWriter::writeWaveFormatProperty(const PROPVARIANT &property, std::wostream &output)
{
    if ((property.vt == VT_EMPTY) || (property.vt == VT_NULL))
    {
        output << "[ NOT SET ]";
        return;
    }

    if (property.vt != VT_BLOB)
    {
        output << "[ Error: Property not a wave format. ]";
        return;
    }

    WAVEFORMATEX *pFmt = reinterpret_cast<WAVEFORMATEX *>(property.blob.pBlobData);
    switch (pFmt->wFormatTag)
    {
        case WAVE_FORMAT_IEEE_FLOAT:
        {
            output << "Floating Point" << std::endl;
            break;
        }
        case WAVE_FORMAT_PCM:
        {
            output << "PCM" << std::endl;
            break;
        }
        case WAVE_FORMAT_EXTENSIBLE:
        {
            const WAVEFORMATEXTENSIBLE *pFormatExtensible =
                reinterpret_cast<const WAVEFORMATEXTENSIBLE *>(property.blob.pBlobData);

            if (pFormatExtensible->SubFormat == KSDATAFORMAT_SUBTYPE_IEEE_FLOAT)
            {
                output << "Floating Point" << std::endl;
            }
            else
            if (pFormatExtensible->SubFormat == KSDATAFORMAT_SUBTYPE_PCM)
            {
                output << "PCM" << std::endl;
            }
            else
            {
                output << "Unknown" << std::endl;
            }

            output << "\tValid Bits Per Sample: " << pFormatExtensible->Samples.wValidBitsPerSample << std::endl;
            break;
        }
        default :
        {
            output << "Unsupported Format for Application" << std::endl;
            break;
        }
    }

    output << "\tChannels: " << pFmt->nChannels << std::endl
           << "\tSamples Per Sec: " << pFmt->nSamplesPerSec << std::endl
           << "\tBlock Align: " << pFmt->nBlockAlign << std::endl
           << "\tBits Per Sample: " << pFmt->wBitsPerSample;
}
