// We have to include InitGuid.h first so that our property IDs will be defined.
#include <InitGuid.h> 

#include "RenderStream.h"
#include "ReferenceTimeUtils.h"
#include "AudioFormat.h"
#include "SampleTypeConverter.h"
#include "SineWaveGenerator.h"

#include <stdexcept>

RenderStream::RenderStream(IMMDevicePtr &spDevice) :
    m_bufferSizeInFrames(0), m_volume(0.05), m_hEvent(NULL)
{
    if (!spDevice)
        throw std::runtime_error("Null device. Can't create audio stream.");

    // Activating the device, which gives us the audio client.
    IAudioClientPtr spAudioClient;
    if (FAILED(spDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL, 
        reinterpret_cast<void **>(&spAudioClient))))
    {
        throw std::runtime_error("Failed to acquire default audio client.");
    }

    // Lets find out what format the device is currently using. This should be 
    // a valid format to use.
    IPropertyStorePtr spPropStore;
    if (FAILED(spDevice->OpenPropertyStore(STGM_READ, &spPropStore)))
    {
        throw std::runtime_error("Couldn't open property store for device!");
    }
    PROPVARIANT varAudioFormat;
    PropVariantInit(&varAudioFormat);
    if (FAILED(spPropStore->GetValue(PKEY_AudioEngine_DeviceFormat, &varAudioFormat)))
    {
        throw std::runtime_error("Couldn't determine current format of the device.");
    }

    std::shared_ptr<AudioFormat> spAudioFormat = std::make_shared<AudioFormat>(
        reinterpret_cast<WAVEFORMATEX *>(varAudioFormat.blob.pBlobData));

    // Determining the audio device's minimum processing period, which helps us
    // choose a reasonable buffer size.
    REFERENCE_TIME audioDeviceMinPeriodInRefTimes = 0;
    if (FAILED(spAudioClient->GetDevicePeriod(NULL, &audioDeviceMinPeriodInRefTimes)))
    {
        throw std::runtime_error("Couldn't determine device period.");
    }

    REFERENCE_TIME bufferLengthInRefTimes = audioDeviceMinPeriodInRefTimes * 2;

    // Allocates two buffers
    HRESULT hr = 0;
    if (FAILED(hr = spAudioClient->Initialize(AUDCLNT_SHAREMODE_EXCLUSIVE, 
        AUDCLNT_STREAMFLAGS_EVENTCALLBACK, bufferLengthInRefTimes, 
        bufferLengthInRefTimes, spAudioFormat->getWaveFormat(), NULL)))
    { 
        // Can occur in Windows 7 and later.
        if (hr == AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED)
            bufferLengthInRefTimes = reinitializeStreamWithAlignedBuffer(spDevice);
        else
            throw std::runtime_error("Could not initialize stream.");
    }

    // Setting up the event handle.
    m_hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (m_hEvent == NULL)
    {
        throw std::runtime_error("Could not create event.");
    }

    if (FAILED(spAudioClient->SetEventHandle(m_hEvent)))
    {
        throw std::runtime_error("Could not set event on audio client.");
    }

    // Obtaining render client.
    IAudioRenderClientPtr spRenderClient;
    if (FAILED(spAudioClient->GetService(__uuidof(IAudioRenderClient),
        reinterpret_cast<void **>(&spRenderClient))))
    {
        throw std::runtime_error("Failed to acquire render client.");
    }

    // Getting total number of frames, which may be different than our ask.
    if (FAILED(spAudioClient->GetBufferSize(&m_bufferSizeInFrames)))
    {
        throw std::runtime_error("Couldn't obtain total buffer frame count.");
    }

    m_spAudioClient = spAudioClient;
    m_spRenderClient = spRenderClient;
    m_spAudioFormat = spAudioFormat;
}

RenderStream::~RenderStream()
{
    if (m_hEvent)
        CloseHandle(m_hEvent);
}

// Called from the render thread. Performs a wait appropriate for our 
// buffering strategy. In this case, we wait for an event to be set.
bool
RenderStream::waitOnBuffer()
{
    if (WaitForSingleObject(m_hEvent, 500) != WAIT_OBJECT_0)
        return false;

    return true;
}

void
RenderStream::setVolume(double volume)
{
    if (volume < 0.0f)
        volume = 0.0f;
    if (volume > 1.0f)
        volume = 1.0f;

    m_volume = volume;
}

void
RenderStream::renderNextFrames(SineWaveGenerator &sampleSource)
{
    // Obtaining a pointer to the available frames.
    BYTE *pBuffer = nullptr;
    if (FAILED(m_spRenderClient->GetBuffer(m_bufferSizeInFrames, &pBuffer)))
    {
        throw std::runtime_error("Couldn't obtain buffer.");
    }

    fillBuffer(sampleSource, pBuffer, m_bufferSizeInFrames);

    // Marks the frames we just wrote as valid.
    m_spRenderClient->ReleaseBuffer(m_bufferSizeInFrames, 0 /*flags*/);
}

void 
RenderStream::fillBuffer(SineWaveGenerator &sampleSource,
    BYTE *pBuffer, UINT32 bufferSizeInFrames)
{
    BYTE *pBegin = pBuffer;
    BYTE *pEnd   = pBuffer + (bufferSizeInFrames * 
        m_spAudioFormat->getFrameSizeInBytes());

    for (; pBegin < pEnd;)
    {
        double currentSample = sampleSource.nextSample(
            m_spAudioFormat->getSampleRateInHz()) * m_volume;

        // We just fill all channels with the same sample,
        for (int i = 0; i < m_spAudioFormat->getNumChannels(); ++i)
        {
            switch (m_spAudioFormat->getEncoding())
            {
                case AudioEncoding::FLOATING_POINT:
                {
                    SampleTypeConverter::doubleToFloat(currentSample, 
                        *reinterpret_cast<float *>(pBegin));
                    pBegin += sizeof(float);
                    break;
                }
                case AudioEncoding::PCM_8:
                {
                    SampleTypeConverter::doubleTo8BitUnsigned(currentSample, 
                        *reinterpret_cast<uint8_t *>(pBegin));
                    pBegin += 1;
                    break;
                }
                case AudioEncoding::PCM_16:
                {
                    SampleTypeConverter::doubleTo16BitSigned(currentSample, 
                        *reinterpret_cast<int16_t *>(pBegin));
                    pBegin += 2;
                    break;
                }
                case AudioEncoding::PCM_24:
                {                    
                    SampleTypeConverter::doubleTo24BitSigned(currentSample, pBegin);
                    pBegin += 3;
                    break;
                }
                case AudioEncoding::PCM_24_IN_32:
                {
                    SampleTypeConverter::doubleTo24BitIn32Signed(currentSample, pBegin);
                    pBegin += 4;
                    break;
                }
                case AudioEncoding::PCM_32:
                {
                    SampleTypeConverter::doubleTo32BitSigned(currentSample, 
                        *reinterpret_cast<int32_t *>(pBegin));
                    pBegin += 4;
                    break;
                }
                default :
                {
                    throw std::runtime_error("Unsupported audio format.");
                }
            }
        }
    }
}

REFERENCE_TIME
RenderStream::reinitializeStreamWithAlignedBuffer(IMMDevicePtr &spDevice)
{
    if (!m_spAudioClient)
    {
        throw std::runtime_error("Can't reinitialize a NULL audio client.");
    }

    // Retrieves the next highest buffer size, in frames.
    UINT32 bufferLengthInFrames = 0;
    if (FAILED(m_spAudioClient->GetBufferSize(&bufferLengthInFrames)))
    {
        throw std::runtime_error("Could not determine appropriate buffer size.");
    }
    m_spAudioClient.Release();

// Calculating an aligned buffer size in REFERENCE_TIME units, per MSDN.
// https://msdn.microsoft.com/en-us/library/windows/desktop/dd370875%28v=vs.85%29.aspx

    REFERENCE_TIME bufferLengthInRefTimes = 
        (REFERENCE_TIME)((10000.0 * 1000 / 
        m_spAudioFormat->getSampleRateInHz() * bufferLengthInFrames) + 0.5);

    // Reactivating the device, which gives us the audio client.
    if (FAILED(spDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL,
        reinterpret_cast<void **>(&m_spAudioClient))))
    {
        throw std::runtime_error("Failed to acquire default audio client.");
    }

    if (FAILED(m_spAudioClient->Initialize(AUDCLNT_SHAREMODE_EXCLUSIVE, 
        AUDCLNT_STREAMFLAGS_EVENTCALLBACK, bufferLengthInRefTimes, 
        bufferLengthInRefTimes, m_spAudioFormat->getWaveFormat(), NULL)))
    {
        throw std::runtime_error(
            "Failed to initialize audio client with aligned buffer size.");
    }

    return bufferLengthInRefTimes;
}
