// We have to include InitGuid.h first so that our property IDs will be defined.
#include <InitGuid.h>

#include "RenderStream.h"
#include "ReferenceTimeUtils.h"
#include "AudioFormat.h"
#include "SampleTypeConverter.h"
#include "SineWaveGenerator.h"

#include <stdexcept>

RenderStream::RenderStream(IMMDevicePtr &spDevice) :
    m_bufferSizeInFrames(0), m_volume(0.05), m_sleepTimeInMs(0)
{
    if (!spDevice)
        throw std::runtime_error("Null device. Can't create audio stream.");

    // Activating the device, which gives us the audio client.
    IAudioClientPtr spAudioClient;
    if (FAILED(spDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL, 
        reinterpret_cast<void **>(&spAudioClient))))
    {
        throw std::runtime_error("Failed to acquire default audio client.");
    }

    // Lets find out what format the device is currently using. This should be 
    // a valid format to use.
    IPropertyStorePtr spPropStore;
    if (FAILED(spDevice->OpenPropertyStore(STGM_READ, &spPropStore)))
    {
        throw std::runtime_error("Couldn't open property store for device!");
    }
    PROPVARIANT varAudioFormat;
    PropVariantInit(&varAudioFormat);
    if (FAILED(spPropStore->GetValue(PKEY_AudioEngine_DeviceFormat, &varAudioFormat)))
    {
        throw std::runtime_error("Couldn't determine current format of the device.");
    }

    std::shared_ptr<AudioFormat> spAudioFormat = std::make_shared<AudioFormat>(
        reinterpret_cast<WAVEFORMATEX *>(varAudioFormat.blob.pBlobData));

    // Determining the audio device's minimum processing period, which helps us
    // choose a reasonable buffer size.
    REFERENCE_TIME audioDeviceMinPeriodInRefTimes = 0;
    if (FAILED(spAudioClient->GetDevicePeriod(NULL, &audioDeviceMinPeriodInRefTimes)))
    {
        throw std::runtime_error("Couldn't determine device period.");
    }
    REFERENCE_TIME bufferLengthInRefTimes = audioDeviceMinPeriodInRefTimes * 2;

    // Initializing audio client.
    if (FAILED(spAudioClient->Initialize(AUDCLNT_SHAREMODE_EXCLUSIVE, 0,
        bufferLengthInRefTimes, 0, spAudioFormat->getWaveFormat(), NULL)))
    { 
        throw std::runtime_error(
            "Could not initialize audio client with exclusive mode.");
    }

    // Obtaining render client.
    IAudioRenderClientPtr spRenderClient;
    if (FAILED(spAudioClient->GetService(__uuidof(IAudioRenderClient),
        reinterpret_cast<void **>(&spRenderClient))))
    {
        throw std::runtime_error("Failed to acquire render client.");
    }

    // Getting total number of frames, which may be different than our ask.
    if (FAILED(spAudioClient->GetBufferSize(&m_bufferSizeInFrames)))
    {
        throw std::runtime_error("Couldn't obtain total buffer frame count.");
    }

    // Determining an appropriate time interval to sleep between
    // filling the data buffer.
    double durationOfBufferInSeconds = static_cast<double>(m_bufferSizeInFrames) /
        static_cast<double>(spAudioFormat->getSampleRateInHz());
    double durationOfBufferInMillis = durationOfBufferInSeconds * 1000.0;
    m_sleepTimeInMs = (DWORD)(durationOfBufferInMillis / 2.0);

    m_spAudioClient = spAudioClient;
    m_spRenderClient = spRenderClient;
    m_spAudioFormat = spAudioFormat;
}

// Called from the render thread. Performs a wait appropriate for our 
// buffering strategy. In this case, we simply sleep for a given duration
// of time. 
bool 
RenderStream::waitOnBuffer()
{
    Sleep(m_sleepTimeInMs);
    return true;
}

void
RenderStream::setVolume(double volume)
{
    if (volume < 0.0f)
        volume = 0.0f;
    if (volume > 1.0f)
        volume = 1.0f;

    m_volume = volume;
}

void
RenderStream::renderNextFrames(SineWaveGenerator &sampleSource)
{
    // See how much buffer space is available. Frame padding is the amount
    // of valid, unread data currently in the buffer. We can use this to figure
    // out how many frames are invalid and available for writing.
    UINT32 numFramesPadding = 0;
    m_spAudioClient->GetCurrentPadding(&numFramesPadding);

    UINT32 numFramesAvailable = m_bufferSizeInFrames - numFramesPadding;

    BYTE *pBuffer = nullptr;
    if (FAILED(m_spRenderClient->GetBuffer(numFramesAvailable, &pBuffer)))
        throw std::runtime_error("Couldn't obtain buffer.");

    fillBuffer(sampleSource, pBuffer, numFramesAvailable);

    // Marks the frames we just wrote as valid.
    m_spRenderClient->ReleaseBuffer(numFramesAvailable, 0 /*flags*/);
}

void 
RenderStream::fillBuffer(SineWaveGenerator &sampleSource,
    BYTE *pBuffer, UINT32 bufferSizeInFrames)
{
    BYTE *pBegin = pBuffer;
    BYTE *pEnd   = pBuffer + (bufferSizeInFrames * 
        m_spAudioFormat->getFrameSizeInBytes());

    for (; pBegin < pEnd;)
    {
        double currentSample = sampleSource.nextSample(
            m_spAudioFormat->getSampleRateInHz()) * m_volume;

        // We just fill all channels with the same sample,
        for (int i = 0; i < m_spAudioFormat->getNumChannels(); ++i)
        {
            switch (m_spAudioFormat->getEncoding())
            {
                case AudioEncoding::FLOATING_POINT:
                {
                    SampleTypeConverter::doubleToFloat(currentSample, 
                        *reinterpret_cast<float *>(pBegin));
                    pBegin += sizeof(float);
                    break;
                }
                case AudioEncoding::PCM_8:
                {
                    SampleTypeConverter::doubleTo8BitUnsigned(currentSample, 
                        *reinterpret_cast<uint8_t *>(pBegin));
                    pBegin += 1;
                    break;
                }
                case AudioEncoding::PCM_16:
                {
                    SampleTypeConverter::doubleTo16BitSigned(currentSample, 
                        *reinterpret_cast<int16_t *>(pBegin));
                    pBegin += 2;
                    break;
                }
                case AudioEncoding::PCM_24:
                {                    
                    SampleTypeConverter::doubleTo24BitSigned(currentSample, pBegin);
                    pBegin += 3;
                    break;
                }
                case AudioEncoding::PCM_24_IN_32:
                {
                    SampleTypeConverter::doubleTo24BitIn32Signed(currentSample, pBegin);
                    pBegin += 4;
                    break;
                }
                case AudioEncoding::PCM_32:
                {
                    SampleTypeConverter::doubleTo32BitSigned(currentSample, 
                        *reinterpret_cast<int32_t *>(pBegin));
                    pBegin += 4;
                    break;
                }
                default :
                {
                    throw std::runtime_error("Unsupported audio format.");
                }
            }
        }
    }
}

