#ifndef _sampletypeconverter__h_
#define _sampletypeconverter__h_

#include <cstdint>
#include <cmath>
#include <windows.h>

#if __BYTE_ORDER != __LITTLE_ENDIAN
#error Unsupported architecture.
#endif

/**
 * Utility class for converting sample types between those used 
 * internally by the application and sample types used by the Windows 
 * audio-engine or audio hardware.
 */
class SampleTypeConverter
{
public:

    // Converts a double value to a float.
    inline static void doubleToFloat(double input, float &output)
    {
        output = static_cast<float>(limit(input, -1.0, 1.0));
    }

    // Converts a float to a double
    inline static void doubleFromFloat(float input, double &output)
    {
        output = static_cast<double>(limit(input, -1.0f, 1.0f));
    }

    // Converts a double value to an unsigned 8-bit value.
    inline static void doubleTo8BitUnsigned(double input, uint8_t &output)
    {
        int16_t adjustedInput = static_cast<int16_t>(input * 128.0);
        adjustedInput += 128;
        adjustedInput = limit(adjustedInput, (int16_t) 0, (int16_t) 255);
        output = static_cast<uint8_t>(adjustedInput);
    }

    // Converts an unsigned 8-bit value to a double.
    inline static void doubleFrom8BitUnsigned(uint8_t input, double &output)
    {
        double adjustedInput = input - 0x80;
        output = adjustedInput / (0x80 - 1.0);
        output = limit(output, -1.0, 1.0);
    }

    // Converts a double value to a 16-bit signed number.
    inline static void doubleTo16BitSigned(double input, int16_t &output)
    {
        int32_t adjustedInput = static_cast<int32_t>(input * 0x8000);
        output = static_cast<int16_t>(limit(adjustedInput, -0x8000, 0x8000 - 1));
    }

    // Converts a 16-bit signed number to a double.
    inline static void doubleFrom16BitSigned(int16_t input, double &output)
    {
        double adjustedInput = static_cast<double>(input) / (0x8000 - 1.0);
        output = limit(adjustedInput, -1.0, 1.0);
    }

    // Converts a double value to a 24-bit number.
    inline static void doubleTo24BitSigned(double input, BYTE *pOutput)
    {
        int32_t adjustedInput = static_cast<int32_t>(input * 0x7FFFFF);
        adjustedInput = limit(adjustedInput, -0x7FFFFF, 0x7FFFFF - 1);

        BYTE *pBytes = reinterpret_cast<BYTE *>(&adjustedInput);
        for (size_t i = 0; i < 3; ++i)
            *pOutput++ = *pBytes++;
    }

    // Converts a 24-bit number to a double.
    inline static void doubleFrom24BitSigned(BYTE *pInput, double &output)
    {
        int32_t adjustedInput = 0;
        if (pInput[2] & 0x80) // We need to sign extend if this is negative.
            adjustedInput = (0xff << 24) | (pInput[2] << 16) | 
                            (pInput[1] << 8) | (pInput[0] << 0);
        else
            adjustedInput = (pInput[2] << 16) | (pInput[1] << 8) | (pInput[0] << 0);
        output = static_cast<double>(adjustedInput) / (0x7FFFFF - 1.0);
        output = limit(output, -1.0, 1.0);
    }

    // Converts a double value to a 24-bit number captured in a 32-bit container.
    inline static void doubleTo24BitIn32Signed(double input, BYTE *pOutput)
    {
        int32_t adjustedInput = static_cast<int32_t>(input * 0x7FFFFF);
        adjustedInput = limit(adjustedInput, -0x7FFFFF, 0x7FFFFF - 1);

// According to...
// https://msdn.microsoft.com/en-us/library/windows/hardware/dn653308%28v=vs.85%29.aspx
// ...if the number of valid bits are less than the actual size of the 
// container (e.g. 24 bits in a 32 bit container), then the PCM data 
// should be packed into the "most significant" portion of the 
// container. The unused bits in the least significant portion should 
// be zeroed out.
        
        *pOutput++ = 0; // Least significant byte for little-ending architectures.
        BYTE *pBytes = reinterpret_cast<BYTE *>(&adjustedInput);
        for (size_t i = 0; i < 3; ++i)
            *pOutput++ = *pBytes++;
    }

    // Converts a 24-bit number captured in a 32-bit container to a double
    inline static void doubleFrom24BitIn32Signed(BYTE *pInput, double &output)
    {

// According to...
// https://msdn.microsoft.com/en-us/library/windows/hardware/dn653308%28v=vs.85%29.aspx
// ...if the number of valid bits are less than the actual size of the 
// container (e.g. 24 bits in a 32 bit container), then the PCM data 
// should be packed into the "most significant" portion of the 
// container. The unused bits in the least significant portion should 
// be zeroed out.

        ++pInput; // Ignore the first byte.

        int32_t adjustedInput = 0;        
        if (pInput[2] & 0x80) // We need to sign extend if this is negative.
            adjustedInput = (0xff << 24) | (pInput[2] << 16) | 
                            (pInput[1] << 8) | (pInput[0] << 0);
        else
            adjustedInput = (pInput[2] << 16) | (pInput[1] << 8) | (pInput[0] << 0);
        output = static_cast<double>(adjustedInput) / (0x7FFFFF - 1.0);
        output = limit(output, -1.0, 1.0);
    }

    // Converts a double value to a 32-bit number.
    inline static void doubleTo32BitSigned(double input, int32_t &output)
    {
        int64_t adjustedInput = static_cast<int64_t>(input * 0x7FFFFFFF);
        output = static_cast<int32_t>(limit(adjustedInput, 
            (int64_t)-0x7FFFFFFF, (int64_t) 0x7FFFFFFF - 1));
    }

    // Converts a 32-bit number to a double.
    inline static void doubleFrom32BitSigned(int32_t input, double &output)
    {
        double adjustedInput = static_cast<double>(input) / (0x7FFFFFFF - 1.0);
        output = limit(adjustedInput, -1.0, 1.0);
    }

private:

    // Constrains a given value to a minimum and maximum.
    template<typename T>
    inline static T limit(T value, T minValue, T maxValue)
    {
        if (value > maxValue)
            return maxValue;
        else
        if (value < minValue)
            return minValue;
        else
            return value;
    }

    // Constructor. NOT IMPLEMENTED.
    SampleTypeConverter() = delete;
};

#endif
