#include "RenderStream.h"
#include "ReferenceTimeUtils.h"
#include "AudioFormat.h"
#include "SampleTypeConverter.h"
#include "SineWaveGenerator.h"

#include <stdexcept>

RenderStream::RenderStream(IMMDevicePtr &spDevice) :
    m_bufferSizeInFrames(0), m_volume(0.05), m_hEvent(NULL)
{
    if (!spDevice)
        throw std::runtime_error("Null device. Can't create audio stream.");

    // Activating the device, which gives us the audio client.
    IAudioClientPtr spAudioClient;
    if (FAILED(spDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL, 
        reinterpret_cast<void **>(&spAudioClient))))
    {
        throw std::runtime_error("Failed to acquire default audio client.");
    }

    // Determining current shared-mode mix format.
    WAVEFORMATEX *pWaveFormat = nullptr;
    if (FAILED(spAudioClient->GetMixFormat(&pWaveFormat)))
    {
        throw std::runtime_error("Failed to determine mix format.");
    }

    std::shared_ptr<AudioFormat> spAudioFormat = 
        std::make_shared<AudioFormat>(pWaveFormat);
    
    // Initializing audio client.
    if (FAILED(spAudioClient->Initialize(AUDCLNT_SHAREMODE_SHARED, 
        AUDCLNT_STREAMFLAGS_EVENTCALLBACK, 0, 0, pWaveFormat, NULL)))
    { 
        throw std::runtime_error(
            "Could not initialize audio client with shared mode.");
    }

    // Setting up the event handle.
    m_hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (m_hEvent == NULL)
    {
        throw std::runtime_error("Could not create event.");
    }

    if (FAILED(spAudioClient->SetEventHandle(m_hEvent)))
    {
        throw std::runtime_error("Could not set event on audio client.");
    }

    // Obtaining render client.
    IAudioRenderClientPtr spRenderClient;
    if (FAILED(spAudioClient->GetService(__uuidof(IAudioRenderClient),
        reinterpret_cast<void **>(&spRenderClient))))
    {
        throw std::runtime_error("Failed to acquire render client.");
    }

    // Getting total number of frames, which may be different than our ask.
    if (FAILED(spAudioClient->GetBufferSize(&m_bufferSizeInFrames)))
    {
        throw std::runtime_error("Couldn't obtain total buffer frame count.");
    }

    m_spAudioClient = spAudioClient;
    m_spRenderClient = spRenderClient;
    m_spAudioFormat = spAudioFormat;
}

RenderStream::~RenderStream()
{
    if (m_hEvent)
        CloseHandle(m_hEvent);
}

// Called from the render thread. Performs a wait appropriate for our 
// buffering strategy. In this case, we wait for an event to be set.
bool
RenderStream::waitOnBuffer()
{
    if (WaitForSingleObject(m_hEvent, 500) != WAIT_OBJECT_0)
        return false;

    return true;
}

void
RenderStream::setVolume(double volume)
{
    if (volume < 0.0f)
        volume = 0.0f;
    if (volume > 1.0f)
        volume = 1.0f;

    m_volume = volume;
}

void
RenderStream::renderNextFrames(SineWaveGenerator &sampleSource)
{
    // See how much buffer space is available. Frame padding is the amount
    // of valid, unread data currently in the buffer. We can use this to figure
    // out how many frames are invalid and available for writing.
    UINT32 numFramesPadding = 0;
    m_spAudioClient->GetCurrentPadding(&numFramesPadding);

    UINT32 numFramesAvailable = m_bufferSizeInFrames - numFramesPadding;

    BYTE *pBuffer = nullptr;
    if (FAILED(m_spRenderClient->GetBuffer(numFramesAvailable, &pBuffer)))
        throw std::runtime_error("Couldn't obtain buffer.");

    fillBuffer(sampleSource, pBuffer, numFramesAvailable);

    // Marks the frames we just wrote as valid.
    m_spRenderClient->ReleaseBuffer(numFramesAvailable, 0 /*flags*/);
}

void 
RenderStream::fillBuffer(SineWaveGenerator &sampleSource,
    BYTE *pBuffer, UINT32 bufferSizeInFrames)
{
    BYTE *pBegin = pBuffer;
    BYTE *pEnd   = pBuffer + (bufferSizeInFrames * 
        m_spAudioFormat->getFrameSizeInBytes());

    for (; pBegin < pEnd;)
    {
        double currentSample = sampleSource.nextSample(
            m_spAudioFormat->getSampleRateInHz()) * m_volume;

        // We just fill all channels with the same sample,
        for (int i = 0; i < m_spAudioFormat->getNumChannels(); ++i)
        {
            switch (m_spAudioFormat->getEncoding())
            {
                case AudioEncoding::FLOATING_POINT:
                {
                    SampleTypeConverter::doubleToFloat(currentSample, 
                        *reinterpret_cast<float *>(pBegin));
                    pBegin += sizeof(float);
                    break;
                }
                case AudioEncoding::PCM_8:
                {
                    SampleTypeConverter::doubleTo8BitUnsigned(currentSample, 
                        *reinterpret_cast<uint8_t *>(pBegin));
                    pBegin += 1;
                    break;
                }
                case AudioEncoding::PCM_16:
                {
                    SampleTypeConverter::doubleTo16BitSigned(currentSample, 
                        *reinterpret_cast<int16_t *>(pBegin));
                    pBegin += 2;
                    break;
                }
                case AudioEncoding::PCM_24:
                {                    
                    SampleTypeConverter::doubleTo24BitSigned(currentSample, pBegin);
                    pBegin += 3;
                    break;
                }
                case AudioEncoding::PCM_24_IN_32:
                {
                    SampleTypeConverter::doubleTo24BitIn32Signed(currentSample, pBegin);
                    pBegin += 4;
                    break;
                }
                case AudioEncoding::PCM_32:
                {
                    SampleTypeConverter::doubleTo32BitSigned(currentSample, 
                        *reinterpret_cast<int32_t *>(pBegin));
                    pBegin += 4;
                    break;
                }
                default :
                {
                    throw std::runtime_error("Unsupported audio format.");
                }
            }
        }
    }
}

