#include "AppStrings.h"
#include "KeyMonitor.h"
#include "RenderStream.h"
#include "RenderThread.h"
#include "AudioDevices.h"
#include "SineWaveGenerator.h"

#include <iostream>

/**
 * Prints out the state of the adjustable parameters.
 */
void 
printAdjustableParameterStates(const RenderStream &stream,
    const SineWaveGenerator &sineWaveGenerator)
{
    std::cout << "Rendering " << sineWaveGenerator.getFrequencyInHz()
              << "Hz with volume of " << stream.getVolume()
              << std::endl;
}

/**
 * Prints out the audio format being used.
 */
void 
printAudioFormat(const AudioFormat &audioFormat)
{
    std::cout << "Audio Format: " << audioFormat.getSampleRateInHz() << " Hz, ";
    switch (audioFormat.getEncoding())
    {
        case AudioEncoding::FLOATING_POINT:
            std::cout << "32-Bit Floating Point"; break;
        case AudioEncoding::PCM_8:
            std::cout << "8-Bit PCM"; break;
        case AudioEncoding::PCM_16:
            std::cout << "16-Bit PCM"; break;
        case AudioEncoding::PCM_24:
            std::cout << "24-Bit PCM"; break;
        case AudioEncoding::PCM_24_IN_32:
            std::cout << "24-Bit PCM (32-bit Container)"; break;
        case AudioEncoding::PCM_32:
            std::cout << "PCM_32 PCM"; break;
    }
    std::cout << std::endl << std::endl;
}

int 
main(int argc, const char * argv[])
{
    SetConsoleTitleA(CONSOLE_TITLE);

    std::ios::sync_with_stdio();

    // Initializing COM. This is required for all applications making use
    // of COM components.
    HRESULT hr = CoInitializeEx(NULL, COINIT_APARTMENTTHREADED);

    std::cout << "Simple Tone Generator using " MODE_TITLE << std::endl
              << "Press 'q' to quit, up / down for volume, "
              << "left / right to change frequency." 
              << std::endl;

    KeyMonitor keyMonitor;
    SineWaveGenerator sineWaveGenerator;

    try
    {
        IMMDevicePtr spDevice = AudioDevices::getDefaultRenderDevice();
        RenderStream stream(spDevice);
        printAudioFormat(*stream.getAudioFormat());
        printAdjustableParameterStates(stream, sineWaveGenerator);

        RenderThread thread(stream, sineWaveGenerator);
        thread.start();
        
        bool quit = false;
        do
        {
            KeyMonitor::KeyPress keyPress = keyMonitor.getKeyPress();
            switch (keyPress)
            {
                case KeyMonitor::KeyPress::Up : // Pitch up
                {
                    stream.setVolume(stream.getVolume() + 0.005f);
                    break;
                }
                case KeyMonitor::KeyPress::Down : // Pitch down
                {
                    stream.setVolume(stream.getVolume() - 0.005f);
                    break;
                }
                case KeyMonitor::KeyPress::Right : // Volume up
                {
                    sineWaveGenerator.setFrequencyInHz(
                        sineWaveGenerator.getFrequencyInHz() + 10.0f);
                    break;
                }
                case KeyMonitor::KeyPress::Left : // Volume down
                {
                    sineWaveGenerator.setFrequencyInHz(
                        sineWaveGenerator.getFrequencyInHz() - 10.0f);
                    break;
                }
                case KeyMonitor::KeyPress::Q:
                {
                    quit = true;
                    break;
                }
            }

            if (keyPress != KeyMonitor::KeyPress::None)
                printAdjustableParameterStates(stream, sineWaveGenerator);

        } while (!quit);

        thread.stop();
    }
    catch (const std::runtime_error &e)
    {
        std::cerr << e.what() << std::endl;
    }    

    CoUninitialize();

    return 0;
}
